/*
								+----------------------------------+
								|                                  |
								|    ***  convex.madness01  ***    |
								|                                  |
								|   Copyright  -tHE SWINe- 2007   |
								|                                  |
								|             Main.cpp             |
								|                                  |
								+----------------------------------+
*/

#ifdef WIN32
#include <windows.h> // gl.h needs some typedefs
#endif
#include <gl/gl.h>
#include <gl/glut.h>
#include <gl/glext.h>
#include <assert.h>
#include <stdio.h>
#include <math.h>
#include <vector>
#include <algorithm>
#include "Vector.h"
#include "OpenGL20.h"
#include "VertexBufferObject.h"
#include "Timer.h"
#include "Torus.h"

//const float f_epsilon = 1e-3f; // in Vector.cpp

struct TMyVertex {
	Vector3f v_position; // position

	Vector3f v_normal;
	Vector4f v_color;
	Vector4f p_texture_coord[4];
	// some more coordinates to make it more curly

	typedef TMyVertex TIntermediate;
	// we will need structure to store intermediate results in.
	// later on, we'll see why is that

	/*
	 *	inline TMyVertex()
	 *		- default constructor, do nothing
	 */
	inline TMyVertex() {};

	/*
	 *	inline TMyVertex(float f_x)
	 *		- constructor with position x parameter
	 *		  (initializes position to [f_x, 0, 0], all other parameters to zero)
	 */
	inline TMyVertex(float f_x)
		:v_position(f_x, 0, 0), v_normal(0, 0, 0), v_color(0, 0, 0, 0)
	{
		for(int i = 0; i < 4; ++ i)
			p_texture_coord[i] = Vector4f(0, 0, 0, 0);
	}

	/*
	 *	inline TMyVertex(float f_x, float f_y, float f_z)
	 *		- constructor with 3-space position parameter
	 *		  (initializes position to [f_x, f_y, f_z], all other parameters to zero)
	 */
	inline TMyVertex(float f_x, float f_y, float f_z)
		:v_position(f_x, f_y, f_z), v_normal(0, 0, 0), v_color(0, 0, 0, 0)
	{
		for(int i = 0; i < 4; ++ i)
			p_texture_coord[i] = Vector4f(0, 0, 0, 0);
	}

	/*
	 *	inline TMyVertex(const Vector3f &r_v_vector)
	 *		- constructor with 3-space position parameter
	 *		  (initializes position to r_v_vector, all other parameters to zero)
	 */
	inline TMyVertex(const Vector3f &r_v_vector)
		:v_position(r_v_vector), v_normal(0, 0, 0), v_color(0, 0, 0, 0)
	{
		for(int i = 0; i < 4; ++ i)
			p_texture_coord[i] = Vector4f(0, 0, 0, 0);
	}

	/*
	 *	inline TMyVertex(const TMyVertex &r_t_vertex)
	 *		- copy-constructor
	 */
	inline TMyVertex(const TMyVertex &r_t_vertex)
		:v_position(r_t_vertex.v_position), v_normal(r_t_vertex.v_normal),
		v_color(r_t_vertex.v_color)
	{
		for(int i = 0; i < 4; ++ i)
			p_texture_coord[i] = r_t_vertex.p_texture_coord[i];
	}

	/*
	 *	inline operator Vector3f() const
	 *		- conversion to Vector3f should return vertex position
	 */
	inline operator Vector3f() const
	{
		return v_position;
	}

	/*
	 *	inline TMyVertex operator =(const Vector3f &r_v_position)
	 *		- assigning Vector3f should write to vertex position
	 */
	inline TMyVertex operator =(const Vector3f &r_v_position)
	{
		v_position = r_v_position;
		return *this;
	}

	/*
	 *	inline TMyVertex operator =(const TMyVertex &r_t_vertex)
	 *		- we need to specify such an assignment as well to avoid converting
	 *		  TMyVertex to Vector3f when assigning to another TMyVertex
	 */
	inline TMyVertex operator =(const TMyVertex &r_t_vertex)
	{
		v_position = r_t_vertex.v_position;
		v_normal = r_t_vertex.v_normal;
		v_color = r_t_vertex.v_color;
		for(int i = 0; i < 4; ++ i)
			p_texture_coord[i] = r_t_vertex.p_texture_coord[i];
		return *this;
	}

	/*
	 *	inline bool operator ==(const TMyVertex &r_t_vertex)
	 *		- we might need means to compare two vertices
	 *		- constant f_epsilon is used as maximal coordinate distance tolerance
	 */
	inline bool operator ==(const TMyVertex &r_t_vertex)
	{
		if((v_position - r_t_vertex.v_position).f_Length() >= f_epsilon ||
		   (v_normal - r_t_vertex.v_normal).f_Length() >= f_epsilon ||
		   (v_color - r_t_vertex.v_color).f_Length() >= f_epsilon)
			return false;
		for(int i = 0; i < 4; ++ i) {
			if((p_texture_coord[i] - r_t_vertex.p_texture_coord[i]).f_Length() >= f_epsilon)
				return false;
		}
		return true;
	}

	/*
	 *	TVertex operator *(float f_scalar) const
	 *		- return vertex, scaled by f_scalar
	 */
	TMyVertex operator *(float f_scalar) const
	{
		TMyVertex t_tmp;

		t_tmp.v_position = v_position * f_scalar;
		t_tmp.v_normal = v_normal * f_scalar;
		t_tmp.v_color = v_color * f_scalar;
		for(int i = 0; i < 4; ++ i)
			t_tmp.p_texture_coord[i] = p_texture_coord[i] * f_scalar;
		// scale vertex

		return t_tmp;
	}

	/*
	 *	TMyVertex operator +(const TMyVertex &r_vertex) const
	 *		- return vertex with sum of coordinates of both original vertices
	 */
	TMyVertex operator +(const TMyVertex &r_vertex) const
	{
		TMyVertex t_tmp;

		t_tmp.v_position = v_position + r_vertex.v_position;
		t_tmp.v_normal = v_normal + r_vertex.v_normal;
		t_tmp.v_color = v_color + r_vertex.v_color;
		for(int i = 0; i < 4; ++ i)
			t_tmp.p_texture_coord[i] = p_texture_coord[i] + r_vertex.p_texture_coord[i];
		// add vertex

		return t_tmp;
	}

	/*
	 *	TVertex operator *=(float f_scalar)
	 *		- return vertex, scaled by f_scalar
	 */
	TMyVertex operator *=(float f_scalar)
	{
		v_position *= f_scalar;
		v_normal *= f_scalar;
		v_color *= f_scalar;
		for(int i = 0; i < 4; ++ i)
			p_texture_coord[i] *= f_scalar;
		// scale vertex

		return *this;
	}

	/*
	 *	TMyVertex operator +=(const TMyVertex &r_vertex)
	 *		- return vertex with sum of coordinates of both original vertices
	 */
	TMyVertex operator +=(const TMyVertex &r_vertex)
	{
		v_position += r_vertex.v_position;
		v_normal += r_vertex.v_normal;
		v_color += r_vertex.v_color;
		for(int i = 0; i < 4; ++ i)
			p_texture_coord[i] += r_vertex.p_texture_coord[i];
		// add vertex

		return *this;
	}
};

struct TRefVertex {
	TMyVertex *m_p_ref; // referenced vertex
	std::vector<TMyVertex*> &m_r_vertex_pool; // vertex pool (here we going to add new vertices)

	typedef TMyVertex TIntermediate;
	// intermediate type for calculations will be TMyVertex

	/*
	 *	inline TRefVertex(const TMyVertex &r_t_vertex, std::vector<TMyVertex*> &r_vertex_pool)
	 *		- default constructor
	 */
	inline TRefVertex(const TMyVertex &r_t_vertex, std::vector<TMyVertex*> &r_vertex_pool)
		:m_p_ref(0), m_r_vertex_pool(r_vertex_pool)
	{
		m_p_ref = new TMyVertex(r_t_vertex); // alloc a new TMyVertex
		m_r_vertex_pool.push_back(m_p_ref); // insert it to the list
	}

	/*
	 *	inline TRefVertex(std::vector<TMyVertex*>::iterator p_vertex,
	 *		std::vector<TMyVertex*> &r_vertex_pool)
	 *		- reference vertex that is already in the list
	 */
	inline TRefVertex(std::vector<TMyVertex*>::iterator p_vertex,
		std::vector<TMyVertex*> &r_vertex_pool)
		:m_p_ref(*p_vertex), m_r_vertex_pool(r_vertex_pool)
	{}

	/*
	 *	inline TRefVertex(const TRefVertex &r_t_vertex)
	 *		- copy constructor
	 */
	inline TRefVertex(const TRefVertex &r_t_vertex)
		:m_p_ref(r_t_vertex.m_p_ref), m_r_vertex_pool(r_t_vertex.m_r_vertex_pool)
	{}

	/*
	 *	inline TRefVertex operator =(const TRefVertex &r_t_vertex)
	 *		- copy operator = (need it to avoid unwanted conversions)
	 */
	inline TRefVertex operator =(const TRefVertex &r_t_vertex)
	{
		m_p_ref = r_t_vertex.m_p_ref;
		m_r_vertex_pool = r_t_vertex.m_r_vertex_pool;

		return *this;
	}

	/*
	 *	inline TRefVertex operator =(const TMyVertex &r_t_vertex)
	 *		- conversion from TMyVertex to TRefVertex (conversion
	 *		  is solved by adding vertex to the vertex list)
	 */
	inline TRefVertex operator =(const TMyVertex &r_t_vertex)
	{
		m_p_ref = new TMyVertex(r_t_vertex); // alloc a new TMyVertex
		m_r_vertex_pool.push_back(m_p_ref); // insert it to the list

		return *this;
	}

	/*
	 *	inline operator TMyVertex() const
	 *		- now we need means to convert vertex type to intermediate vertex type
	 */
	inline operator TMyVertex() const
	{
		return *m_p_ref;
	}

	/*
	 *	inline operator Vector3f() const
	 *		- conversion to Vector3f should return vertex position
	 */
	inline operator Vector3f() const
	{
		return m_p_ref->v_position;
	}

	/*
	 *	inline TRefVertex operator =(const Vector3f &r_v_position)
	 *		- assigning Vector3f should write to vertex position
	 */
	inline TRefVertex operator =(const Vector3f &r_v_position)
	{
		m_p_ref->v_position = r_v_position;
		return *this;
	}

	/*
	 *	inline bool operator ==(const TMyVertex &r_t_vertex)
	 *		- we might need means to compare two vertices
	 *		- constant f_epsilon is used as maximal coordinate distance tolerance
	 */
	inline bool operator ==(const TMyVertex &r_t_vertex)
	{
		return *m_p_ref == r_t_vertex;
	}

	/*
	 *	TVertex operator *(float f_scalar) const
	 *		- return vertex, scaled by f_scalar
	 *		- returns a new TMyVertex because new vertex originated
	 *		  (in case it will need to be converted to TRefVertex, it will be added to the list)
	 */
	TMyVertex operator *(float f_scalar) const
	{
		return *m_p_ref * f_scalar;
	}

	/*
	 *	TMyVertex operator +(const TMyVertex &r_vertex) const
	 *		- return vertex with sum of coordinates of both original vertices
	 *		- returns a new TMyVertex because new vertex originated
	 *		  (in case it will need to be converted to TRefVertex, it will be added to the list)
	 */
	TMyVertex operator +(const TMyVertex &r_vertex) const
	{
		return *m_p_ref + r_vertex;
	}

	/*
	 *	TRefVertex operator *=(float f_scalar) const
	 *		- return vertex, scaled by f_scalar
	 *		- returns this because no new vertex originated
	 *		  (coordinates of some existing vertex were scaled only)
	 */
	TRefVertex operator *=(float f_scalar)
	{
		*m_p_ref *= f_scalar;
		return *this;
	}

	/*
	 *	TRefVertex operator +=(const TMyVertex &r_vertex) const
	 *		- return vertex with sum of coordinates of both original vertices
	 *		- returns this because no new vertex originated
	 *		  (coordinates of some existing vertex were scaled only)
	 */
	TRefVertex operator +=(const TMyVertex &r_vertex)
	{
		*m_p_ref += r_vertex;
		return *this;
	}
};

template <class TVertStruct>
class CPolygon2 {
protected:
	std::vector<TVertStruct> m_vertex_list;
	Plane3f m_t_normal;

public:
	/*
	 *	void Delete()
	 *		- delete all polygon vertices
	 */
	void Delete()
	{
		m_vertex_list.clear();
	}

	/*
	 *	inline bool Add_Vertex(const TVertStruct &r_t_vertex)
	 *		- add single vertex past the last vertex in the array
	 */
	inline bool Add_Vertex(const TVertStruct &r_t_vertex)
	{
		return Add_Vertex(m_vertex_list.size(), &r_t_vertex, 1);
	}

	/*
	 *	inline bool Add_Vertex(int n_insert_before, const TVertStruct &r_t_vertex)
	 *		- add single vertex before the n_insert_before-th vertex in the array
	 */
	inline bool Add_Vertex(int n_insert_before, const TVertStruct &r_t_vertex)
	{
		return Add_Vertex(n_insert_before, &r_t_vertex, 1);
	}

	/*
	 *	inline bool Add_Vertex(const TVertStruct *p_vertex, int n_count)
	 *		- add array of vertices past the last vertex in the array
	 */
	inline bool Add_Vertex(const TVertStruct *p_vertex, int n_count)
	{
		return Add_Vertex(m_vertex_list.size(), p_vertex, n_count);
	}

	/*
	 *	bool Add_Vertex(int n_insert_before, const TVertStruct *p_vertex, int n_count)
	 *		- add array of vertices before the n_insert_before-th vertex in the array
	 */
	bool Add_Vertex(int n_insert_before, const TVertStruct *p_vertex, int n_count)
	{
		assert(n_count > 0);
		assert(n_insert_before >= 0 && n_insert_before <= m_vertex_list.size());
		m_vertex_list.reserve(m_vertex_list.size() + n_count);
		if(m_vertex_list.capacity() < m_vertex_list.size() + n_count)
			return false;
		m_vertex_list.insert(m_vertex_list.begin() + n_insert_before,
			p_vertex, p_vertex + n_count);
		return true;
	}

	/*
	 *	void Delete_Vertices(int n_index, int n_count)
	 *		- delete n_count vertices, beggining with vertex n_index
	 */
	void Delete_Vertices(int n_index, int n_count)
	{
		m_vertex_list.erase(&m_vertex_list[n_index], &m_vertex_list[n_index + n_count]);
	}

	/*
	 *	inline int n_Vertex_Num() const
	 *		- return number of vertices
	 */
	inline int n_Vertex_Num() const
	{
		return m_vertex_list.size();
	}

	/*
	 *	inline TVertStruct &r_t_Vertex(int n_index) const
	 *		- vertex access function
	 *		- doesn't check array bounds
	 */
	inline TVertStruct &r_t_Vertex(int n_index)
	{
		return m_vertex_list[n_index];
	}

	/*
	 *	inline TVertStruct &t_Vertex(int n_index) const
	 *		- vertex access function
	 *		- doesn't check array bounds
	 */
	inline const TVertStruct &t_Vertex(int n_index) const
	{
		return m_vertex_list[n_index];
	}

	/*
	 *	const Vector3f v_Center() const
	 *		- return position of center of polygon
	 *		- if polygon has no vertices, return the O vector
	 */
	const Vector3f v_Center() const
	{
		if(!m_vertex_list.size())
			return Vector3f(0, 0, 0);

		Vector3f v_center(0, 0, 0);
		for(int i = 0; i < m_vertex_list.size(); ++ i)
			v_center += m_vertex_list[i];

		return v_center * (1.0f / m_vertex_list.size());
	}

	/*
	 *	const TVertStruct::TIntermediate v_Center2() const
	 *		- return vertex in center of polygon
	 *		- if polygon has no vertices, return the O vertex
	 */
	const TVertStruct::TIntermediate v_Center2() const
	{
		TVertStruct::TIntermediate t_center(Vector3f(0, 0, 0));
		if(!m_vertex_list.size())
			return t_center;

		for(int i = 0; i < m_vertex_list.size(); ++ i)
			t_center += m_vertex_list[i];

		return t_center * (1.0f / m_vertex_list.size());
	}

	/*
	 *	bool Calc_Normal()
	 *		- calculate normal, return false if there wasn't trinity
	 *		  of vertices to construct plane from
	 *		- it has to be called explicitly, it is never called by CPolygon2 itself
	 */
	bool Calc_Normal()
	{
		if(m_vertex_list.size() < 3)
			return false;
		// impossible with so little vertices

		for(int i = 1; i < m_vertex_list.size(); ++ i) {
			Vector3f v_u = (Vector3f)m_vertex_list[i - 1] - (Vector3f)m_vertex_list[i];
			if(v_u.f_Length() < f_epsilon)
				continue;
			// get non-zero vector

			for(int j = 0; j < m_vertex_list.size(); ++ j) {
				if(j == i || j == i - 1)
					continue;
				Vector3f v_v = (Vector3f)m_vertex_list[j] - (Vector3f)m_vertex_list[i];
				if(v_v.f_Length() < f_epsilon)
					continue;
				// get another different non-zero vector

				Vector3f v_normal = v_u.v_Cross(v_v);
				if(v_normal.f_Length() < f_epsilon)
					continue;
				v_normal.Normalize();
				// colinear vectors would yield zero normal

				m_t_normal = Plane3f(m_vertex_list[0], v_normal);
				return true;
				// create plane
			}
		}

		return false;
	}

	/*
	 *	inline const Plane3f &t_Normal() const
	 *		- normal access function
	 */
	inline const Plane3f &t_Normal() const
	{
		return m_t_normal;
	}
};

class CVertexArrays {
protected:
	unsigned int n_vertex_num;
	unsigned int n_vertex_size;
	GLenum n_vertex_format;
	float *p_vertex_array;

	unsigned int n_index_num;
	unsigned int *p_index_array;

	unsigned int n_polygon_num;
	GLsizei *p_vertex_count; // vertex counts of individual polygons
	unsigned int **p_vertex_pointer; // pointers to vertex indices in p_index_array

	CVertexBufferObject vertex_buffer_object, index_buffer_object;
	bool b_use_vbos;
	// buffer objects and flag wheter we can use them

	/*
	 *	class CCountVertices
	 *		- small function object for counting vertices
	 */
	class CCountVertices {
	protected:
		unsigned int m_n_count;

	public:
		inline CCountVertices()
			:m_n_count(0)
		{}

		inline void operator ()(const CPolygon2<TRefVertex> &r_polygon)
		{
			m_n_count += r_polygon.n_Vertex_Num();
		}

		inline operator unsigned int() const
		{
			return m_n_count;
		}
	};

	/*
	 *	class CFillVertexArray
	 *		- small function object we will use to fill up vertex array
	 */
	class CFillVertexArray {
	protected:
		float *m_p_vertex_array;

	public:
		inline CFillVertexArray(float *p_vertex_array)
			:m_p_vertex_array(p_vertex_array)
		{}

		inline void operator ()(const TMyVertex *p_vertex)
		{
			m_p_vertex_array[0] = p_vertex->v_normal.x;
			m_p_vertex_array[1] = p_vertex->v_normal.y;
			m_p_vertex_array[2] = p_vertex->v_normal.z;
			m_p_vertex_array[3] = p_vertex->v_position.x;
			m_p_vertex_array[4] = p_vertex->v_position.y;
			m_p_vertex_array[5] = p_vertex->v_position.z;
			m_p_vertex_array += 6;
		}
	};

	/*
	 *	class CFillIndexArray
	 *		- small function object we will use to fill up index and index range arrays
	 */
	class CFillIndexArray {
	protected:
		const std::vector<TMyVertex*> &m_r_global_vertex_list;
		unsigned int *m_p_index_array;
		GLsizei *m_p_vertex_count;
		unsigned int **m_p_vertex_pointer;

	public:
		inline CFillIndexArray(const std::vector<TMyVertex*> &r_global_vertex_list,
			unsigned int *p_index_array, GLsizei *p_vertex_count,
			unsigned int **p_vertex_pointer)
			:m_r_global_vertex_list(r_global_vertex_list), m_p_index_array(p_index_array),
			m_p_vertex_count(p_vertex_count), m_p_vertex_pointer(p_vertex_pointer)
		{}

		inline void operator ()(const CPolygon2<TRefVertex> &r_polygon)
		{
			*m_p_vertex_count ++ = r_polygon.n_Vertex_Num();
			*m_p_vertex_pointer ++ = m_p_index_array;
			// add a single array range record

			for(int i = 0, n = r_polygon.n_Vertex_Num(); i < n; ++ i, ++ m_p_index_array) {
				unsigned int n_index = std::find(m_r_global_vertex_list.begin(),
					m_r_global_vertex_list.end(), r_polygon.t_Vertex(i).m_p_ref) -
					m_r_global_vertex_list.begin();
				// find vertex in array, calculate index

				*m_p_index_array = n_index;
			}
			// fill index array
		}
	};

public:
	/*
	 *	CVertexArrays(std::vector<CPolygon2<TRefVertex> >::const_iterator p_begin,
	 *		std::vector<CPolygon2<TRefVertex> >::const_iterator p_end,
	 *		const std::vector<TMyVertex*> &r_global_vertex_list)
	 *		- default constructor
	 *		- p_begin and p_end are two iterators from list of polygons, delimiting
	 *		  first and last polygon to be added to the array
	 *		- r_global_vertex_list is reference to global vertex list
	 *		  (coordinates will be taken out of there)
	 */
	CVertexArrays(std::vector<CPolygon2<TRefVertex> >::const_iterator p_begin,
		std::vector<CPolygon2<TRefVertex> >::const_iterator p_end,
		const std::vector<TMyVertex*> &r_global_vertex_list)
		:p_vertex_array(0), p_index_array(0), p_vertex_count(0), p_vertex_pointer(0)
	{
		n_vertex_format = GL_N3F_V3F; // just three floats for normal and three for position
		n_vertex_size = 6; // ... that makes six floats

		n_vertex_num = r_global_vertex_list.size();
		n_index_num = std::for_each(p_begin, p_end, CCountVertices()); // count indices
		n_polygon_num = p_end - p_begin;

		if(!(p_vertex_array = new float[n_vertex_num * n_vertex_size]) ||
		   !(p_index_array = new unsigned int[n_index_num]) ||
		   !(p_vertex_count = new GLsizei[n_polygon_num]) ||
		   !(p_vertex_pointer = new unsigned int*[n_polygon_num]))
			return;
		// alloc arrays

		std::for_each(r_global_vertex_list.begin(), r_global_vertex_list.end(),
			CFillVertexArray(p_vertex_array));
		// fill vertex array with coordinates

		std::for_each(p_begin, p_end, CFillIndexArray(r_global_vertex_list,
			p_index_array, p_vertex_count, p_vertex_pointer));
		// fill index array and array of index ranges

		if(CVertexBufferObject::b_Supported()) {
			if(!vertex_buffer_object.Alloc(n_vertex_num * n_vertex_size * sizeof(float),
			   p_vertex_array, GL_STATIC_DRAW_ARB) || !index_buffer_object.Alloc(n_index_num *
			   sizeof(unsigned int), p_index_array, GL_STATIC_DRAW_ARB)) {
				fprintf(stderr, "error: not enough memory for vertex buffer objects\n");
				b_use_vbos = false;
				CBufferObjectUnBinder::UnBind();
				return;
			}
			// create buffer objects

			CBufferObjectUnBinder::UnBind();
			// unbind buffers for now

			for(int i = 0; i < n_polygon_num; ++ i)
				p_vertex_pointer[i] = (unsigned int*)0 + (p_vertex_pointer[i] - p_index_array);
			// change pointer base from p_index_array to null

			printf("using vertex buffer objects\n");
			b_use_vbos = true;
		} else {
			fprintf(stderr, "error: vertex buffer objects not supported\n");
			b_use_vbos = false;
		}
	}

	/*
	 *	~CVertexArrays()
	 *		- default destructor
	 */
	~CVertexArrays()
	{
		if(p_vertex_array)
			delete[] p_vertex_array;
		if(p_index_array)
			delete[] p_index_array;
		if(p_vertex_count)
			delete[] p_vertex_count;
		if(p_vertex_pointer)
			delete[] p_vertex_pointer;
	}

	/*
	 *	bool b_Ready() const
	 *		- returns true in case constructor succeeded in creating
	 *		  all the neccesary arrays, otherwise returns false
	 */
	bool b_Ready() const
	{
		return p_vertex_array && p_index_array && p_vertex_count && p_vertex_pointer;
	}

	/*
	 *	void Render()
	 *		- render geometry using vertex arrays
	 */
	void Render()
	{
		glEnableClientState(GL_VERTEX_ARRAY);
		glEnableClientState(GL_NORMAL_ARRAY);
		// those things will be taken from vertex array
		// instead of glVertex3f / glNormal3f

		if(!b_use_vbos) { // non-VBO fallback
			glInterleavedArrays(n_vertex_format, n_vertex_size * sizeof(float), p_vertex_array);
			// give OpenGL pointer to our vertices

			glMultiDrawElements(GL_POLYGON, p_vertex_count,
				GL_UNSIGNED_INT, (const void**)p_vertex_pointer, n_polygon_num);
			// draw all the polygons using a single call
		} else { // fast VBO branch
			vertex_buffer_object.Bind();
			index_buffer_object.BindAs(GL_ELEMENT_ARRAY_BUFFER);
			// bind buffers

			glInterleavedArrays(n_vertex_format, n_vertex_size * sizeof(float), 0);
			// give OpenGL pointer to our vertices
			// note buffer objects use zero pointers

			glMultiDrawElements(GL_POLYGON, p_vertex_count,
				GL_UNSIGNED_INT, (const void**)p_vertex_pointer, n_polygon_num);
			// draw all the polygons using a single call
			
			CBufferObjectUnBinder::UnBind();
			CBufferObjectUnBinder::UnBind(GL_ELEMENT_ARRAY_BUFFER);
			// unbind buffers for now
		}

		glDisableClientState(GL_VERTEX_ARRAY);
		glDisableClientState(GL_NORMAL_ARRAY);
	}
};

int n_width = 640, n_height = 480;

void onResize(int n_new_width, int n_new_height)
{
	glViewport(0, 0, n_width = n_new_width, n_height = n_new_height);
	// set new viewport

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(90, (float)n_width / n_height, .5f, 1000);
	// set perspective projection

	glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	// wireframe

	glMatrixMode(GL_MODELVIEW);
}

CVertexArrays *p_vertex_arrays;
CTimer timer;

void onDisplay()
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glLoadIdentity();
	gluLookAt(120 * sin(timer.f_Time()), -50,
		120 * cos(timer.f_Time()), 0, 0, 0, 0, 1, 0);
	// set camera

	glColor3f(1, 1, 1);
	// white

	if(p_vertex_arrays && p_vertex_arrays->b_Ready())
		p_vertex_arrays->Render();
	// draw our object

	glutSwapBuffers();
}

int main(int n_arg_num, /*const*/ char **p_arg_list) // glut doesn't like 'const'
{
	{
		CPolygon2<TMyVertex> t_my_polygon;

		t_my_polygon.Add_Vertex(TMyVertex(-1, -1, 0));
		t_my_polygon.Add_Vertex(TMyVertex( 1, -1, 0));
		t_my_polygon.Add_Vertex(TMyVertex( 1,  1, 0));
		t_my_polygon.Add_Vertex(TMyVertex(-1,  1, 0));
		// add some vertices

		Vector3f v_center = t_my_polygon.v_Center();
		TMyVertex t_center = t_my_polygon.v_Center2();
		// get center position and center vertex

		t_my_polygon.Calc_Normal();
		Vector3f v_normal = t_my_polygon.t_Normal().v_normal;
		// calculate normal
	}
	// test with individual vertex list inside polygon

	glutInit(&n_arg_num, p_arg_list);
	glutInitDisplayMode(GLUT_RGBA | GLUT_DEPTH | GLUT_DOUBLE);
	glutInitWindowSize(n_width, n_height);
	glutCreateWindow("convex.madness");

	glutDisplayFunc(onDisplay);
	glutIdleFunc(onDisplay); // want to redisplay again and again
	glutReshapeFunc(onResize);
	// init OpenGL using GLUT (easy, huh?)

	{
		CPolygon2<TRefVertex> t_my_polygon;
		std::vector<TMyVertex*> global_vertex_list;

		t_my_polygon.Add_Vertex(TRefVertex(TMyVertex(-1, -1, 0), global_vertex_list));
		t_my_polygon.Add_Vertex(TRefVertex(TMyVertex( 1, -1, 0), global_vertex_list));
		t_my_polygon.Add_Vertex(TRefVertex(TMyVertex( 1,  1, 0), global_vertex_list));
		t_my_polygon.Add_Vertex(TRefVertex(TMyVertex(-1,  1, 0), global_vertex_list));
		// add some vertices

		Vector3f v_center = t_my_polygon.v_Center();
		TRefVertex t_center(t_my_polygon.v_Center2(), global_vertex_list);
		// get center position and center vertex

		t_my_polygon.Calc_Normal();
		Vector3f v_normal = t_my_polygon.t_Normal().v_normal;
		// calculate normal

		std::vector<CPolygon2<TRefVertex> > poly_list;
		poly_list.push_back(t_my_polygon);
		// simple poly-list

		unsigned int n_offset = global_vertex_list.size();
		global_vertex_list.reserve(n_offset + CTorus::n_Vertex_Num());
		if(global_vertex_list.capacity() < n_offset + CTorus::n_Vertex_Num()) {
			fprintf(stderr, "error: not enough memory\n");
			return -1;
		}
		for(unsigned int i = 0; i < CTorus::n_Vertex_Num(); ++ i) {
			TMyVertex *p_vertex;
			if(!(p_vertex = new TMyVertex(CTorus::t_Vertex(i)))) {
				fprintf(stderr, "error: not enough memory\n");
				return -1;
			}
			global_vertex_list.push_back(p_vertex);
		}
		// copy vertex list

		poly_list.reserve(CTorus::n_Face_Num());
		if(poly_list.capacity() < CTorus::n_Face_Num()) {
			fprintf(stderr, "error: not enough memory\n");
			return -1;
		}
		for(unsigned int i = 0; i < CTorus::n_Face_Num(); ++ i) {
			CPolygon2<TRefVertex> t_my_polygon;
			for(int n = 0; n < 3; ++ n) {
				if(!(t_my_polygon.Add_Vertex(TRefVertex(global_vertex_list.begin() +
				   (n_offset + CTorus::n_VertexIndex(i, n)), global_vertex_list)))) {
					fprintf(stderr, "error: not enough memory\n");
					return -1;
				}
			}
			poly_list.push_back(t_my_polygon);
		}
		// copy poly list

		if(!CGLExtensionHandler::n_GetGL14FuncPointers()) {
			p_vertex_arrays = new CVertexArrays(poly_list.begin(),
				poly_list.end(), global_vertex_list);
		} else {
			fprintf(stderr, "error: OpenGL 1.4 not available, don't have vertex arrays\n");
			return -1;
		}
		// create vertex arrays
	}
	// test with global vertex list

	glutMainLoop();
	// begin rendering

	return 0;
}
