/*
								+----------------------------------+
								|                                  |
								|     ***  VBO wrap class  ***     |
								|                                  |
								|   Copyright  -tHE SWINe- 2005   |
								|                                  |
								|      VertexBufferObject.cpp      |
								|                                  |
								+----------------------------------+
*/

/*
 *	passed code revision at 08-13-2006
 *
 *	integers, contanining just true or false were changed to bool
 *	added CPixelBufferObject class
 *	changed behavior of CVertexBufferObject::UnMap() so it returns true when buffer was not mapped
 *	(thereby return walue false always mean buffer data were lost)
 *	renamed CVBOUnBinder to CBufferObjectUnBinder
 *
 */

#ifdef WIN32
#include <windows.h>
#endif
#include <assert.h>
#include <vector>
#include <time.h>
#include <gl/gl.h>

#include <gl/glext.h>
#include "OpenGL20.h"

#include "VertexBufferObject.h"

#define __FuncGuard(x)
#define _ASSERTE(x) assert(x)
// rip-off function guard

/*
 *								=== CVertexBufferObject ===
 */

/*
 *	CVertexBufferObject::CVertexBufferObject(GLenum n_target = GL_ARRAY_BUFFER_ARB)
 *		- default constructor
 *		- valid parameters are GL_ARRAY_BUFFER_ARB (vertex data) and
 *		  GL_ELEMENT_ARRAY_BUFFER_ARB (vertex indices)
 */
CVertexBufferObject::CVertexBufferObject(GLenum n_target/* = GL_ARRAY_BUFFER_ARB*/)
	:m_n_primary_target(n_target), m_n_buffer_id(0),
	m_b_mapped(false), m_n_size(0), m_n_usage(0/*GL_STATIC_DRAW_ARB*/)
{
	__FuncGuard("CVertexBufferObject::CVertexBufferObject");

	CGLExtensionHandler::n_GetVertexBufferObjectARBFuncPointers();
}

/*
 *	CVertexBufferObject::~CVertexBufferObject()
 *		- default destructor
 */
CVertexBufferObject::~CVertexBufferObject()
{
	__FuncGuard("CVertexBufferObject::~CVertexBufferObject");

	if(m_n_buffer_id) {
		if(m_b_mapped)
			UnMap();
		glDeleteBuffersARB(1, &m_n_buffer_id);
		m_n_buffer_id = 0;
		m_n_size = 0;
		m_n_usage = 0; //GL_STATIC_DRAW_ARB?
	}
}

/*
 *	static bool CVertexBufferObject::b_Supported()
 *		- returns true when GL_ARB_VERTEX_BUFFER_OBJECT extension is supported
 *		  and all entry points are located
 */
bool CVertexBufferObject::b_Supported()
{
	__FuncGuard("CVertexBufferObject::b_Supported");

	return CGLExtensionHandler::b_SupportedExtension("GL_ARB_vertex_buffer_object") &&
		!CGLExtensionHandler::n_GetVertexBufferObjectARBFuncPointers();
}

/*
 *	bool CVertexBufferObject::Alloc(int n_size, const void *p_data = NULL,
 *		GLenum n_usage = GL_STATIC_DRAW_ARB)
 *		- alloc buffer data of size n_size bytes, can be initialized to p_data in case
 *		  non-NULL pointer is supplied
 *		- n_usage is optimalization hint only, it is one of:
 *		  GL_STREAM_DRAW_ARB, GL_STREAM_READ_ARB, GL_STREAM_COPY_ARB
 *		  GL_STATIC_DRAW_ARB, GL_STATIC_READ_ARB, GL_STATIC_COPY_ARB
 *		  GL_DYNAMIC_DRAW_ARB, GL_DYNAMIC_READ_ARB, GL_DYNAMIC_COPY_ARB
 *		- in case buffer was allocated already, it's data are freed as soon as there are no
 *		  OpenGL commands, sourcing data from it (it shouldn't block program execution,
 *		  depends on graphics card vendor and driver version) and new buffer is allocated
 *		- it always returns true, but it may generate GL_OUT_OF_MEMORY error (call glGetError())
 */
bool CVertexBufferObject::Alloc(int n_size, const void *p_data/*= NULL*/, GLenum n_usage)
{
	__FuncGuard("CVertexBufferObject::Alloc");

	if(!m_n_buffer_id) {
		glGenBuffersARB(1, &m_n_buffer_id);
		if(!m_n_buffer_id)
			return false;
		glBindBufferARB(m_n_primary_target, m_n_buffer_id);
	}
	// called first time and buffer is not yet allocated ... alloc a new buffer as primary target

	glBufferDataARB(m_n_primary_target, n_size, p_data, n_usage);

	m_n_size = n_size;
	m_n_usage = n_usage;

	return true; // return value is for "software" (non-opengl accelerated)
	// implementations only that may actually alloc something
}

/*
 *	bool CVertexBufferObject::Bind()
 *		- binds the buffer with target set to target, passed to constructor
 *		  (when called first time, it creates OpenGL object as well)
 *		- returns true on success, false on failure
 */
bool CVertexBufferObject::Bind()
{
	__FuncGuard("CVertexBufferObject::Bind");

	if(!m_n_buffer_id) {
		glGenBuffersARB(1, &m_n_buffer_id);
		if(!m_n_buffer_id)
			return false;
	}

	glBindBufferARB(m_n_primary_target, m_n_buffer_id);

	return true;
}

/*
 *	bool CVertexBufferObject::BindAs(GLenum n_target = GL_PIXEL_PACK_BUFFER_ARB)
 *		- binds the buffer with target n_target
 *		  (when called first time, it creates OpenGL object as well)
 *		- returns true on success, false on failure
 *		- valid parameters are GL_ARRAY_BUFFER_ARB (vertex data) and
 *		  GL_ELEMENT_ARRAY_BUFFER_ARB (vertex indices)
 *		- note GL_PIXEL_PACK_BUFFER_ARB and GL_PIXEL_UNPACK_BUFFER_ARB are possible, but
 *		  should only be used with CPixelBufferObject (it's essentialy the same class, but it
 *		  makes you sure you check hardware support of this extension)
 *		- you can copy data using glReadPixels or glGetTexImage with <data> parameter,
 *		  returned by call to p_OffsetPointer(0)
 */
bool CVertexBufferObject::BindAs(GLenum n_target/*= GL_PIXEL_PACK_BUFFER_ARB*/)
{
	__FuncGuard("CVertexBufferObject::BindAs");

	if(!m_n_buffer_id) {
		glGenBuffersARB(1, &m_n_buffer_id);
		if(!m_n_buffer_id)
			return false;
	}

	glBindBufferARB(n_target, m_n_buffer_id);

	return true;
}

/*
 *	void *CVertexBufferObject::p_OffsetPointer(int n_offset)
 *		- returns offset pointer for use with glDrawElements, etc ... (always, can never fail)
 *		- do not use this pointer to access data, it doesn't point to client address space
 */
void *CVertexBufferObject::p_OffsetPointer(int n_offset)
{
	__FuncGuard("CVertexBufferObject::p_OffsetPointer");

	return (void*)((char*)NULL + n_offset);
}

/*
 *	void CVertexBufferObject::BufferSubData(int n_offset, int n_size, const void *p_data)
 *		- re-specify part (or all) of buffer data from p_data
 *		- n_offset must be greater or equal to 0, n_offset + n_size must
 *		  be less or equal to buffer size, otherwise OpenGL error is generated (glGetError())
 */
void CVertexBufferObject::BufferSubData(int n_offset, int n_size, const void *p_data)
{
	__FuncGuard("CVertexBufferObject::BufferSubData");

	glBufferSubDataARB(m_n_primary_target, n_offset, n_size, p_data);
}

/*
 *	void CVertexBufferObject::GetBufferSubData(int n_offset, int n_size, void *p_data)
 *		- read part (or all) of buffer data into p_data
 *		- n_offset must be greater or equal to 0, n_offset + n_size must
 *		  be less or equal to buffer size, otherwise OpenGL error is generated (glGetError())
 */
void CVertexBufferObject::GetBufferSubData(int n_offset, int n_size, void *p_data)
{
	__FuncGuard("CVertexBufferObject::GetBufferSubData");

	glGetBufferSubDataARB(m_n_primary_target, n_offset, n_size, p_data);
}

/*
 *	void *CVertexBufferObject::p_Map(int n_access)
 *		- map buffer into client address space (one buffer only can be mapped at a time)
 *		- n_access is buffer access, it can be one of:
 *		  GL_READ_ONLY_ARB, GL_WRITE_ONLY_ARB, GL_READ_WRITE_ARB
 *		- return client address space pointer in case of success or NULL on failure
 */
void *CVertexBufferObject::p_Map(int n_access)
{
	__FuncGuard("CVertexBufferObject::p_Map");

	void *p_result = glMapBufferARB(m_n_primary_target, n_access);

	if(p_result != NULL) {
		m_n_access = n_access;
		m_b_mapped = true;
	}

	return p_result;
}

/*
 *	void *CVertexBufferObject::p_BufferPointer()
 *		- in case buffer object has been successfuly mapped, return client-space pointer
 *		  otherwise return NULL
 */
void *CVertexBufferObject::p_BufferPointer()
{
	__FuncGuard("CVertexBufferObject::p_BufferPointer");

	if(!m_b_mapped) {
		_ASSERTE(!b_Mapped()); // check the ogl state against m_b_mapped
		return 0;
	}

	void *p_ptr;
	glGetBufferPointervARB(m_n_primary_target, GL_BUFFER_MAP_POINTER_ARB, &p_ptr);

	return p_ptr;
} // return pointer only in case it's been mapped

/*
 *	bool CVertexBufferObject::UnMap()
 *		- unmaps buffer, returns true on success or in case buffer was not mapped
 *		- returns false on failure (buffer contents were lost, may be caused by
 *		  screen-mode switch while buffer was bound)
 */
bool CVertexBufferObject::UnMap()
{
	__FuncGuard("CVertexBufferObject::UnMap");

	if(!m_b_mapped) {
		_ASSERTE(!b_Mapped()); // check the ogl state against m_b_mapped
		return true; // !!
	}

	m_b_mapped = false;

	return (glUnmapBufferARB(m_n_primary_target) == GL_TRUE);
}

/*
 *	int CVertexBufferObject::n_Size()
 *		- returns size of buffer in bytes
 */
int CVertexBufferObject::n_Size()
{
	__FuncGuard("CVertexBufferObject::n_Size");

#ifdef _DEBUG
	int n_size = m_n_size; // must be bound
	glGetBufferParameterivARB(m_n_primary_target, GL_BUFFER_SIZE_ARB, &n_size);
	_ASSERTE(n_size == m_n_size);
#endif

	return m_n_size;
}

/*
 *	int CVertexBufferObject::n_Usage()
 *		- returns buffer ussage; one of:
 *		  GL_STREAM_DRAW_ARB, GL_STREAM_READ_ARB, GL_STREAM_COPY_ARB
 *		  GL_STATIC_DRAW_ARB, GL_STATIC_READ_ARB, GL_STATIC_COPY_ARB
 *		  GL_DYNAMIC_DRAW_ARB, GL_DYNAMIC_READ_ARB, GL_DYNAMIC_COPY_ARB
 */
int CVertexBufferObject::n_Usage()
{
	__FuncGuard("CVertexBufferObject::n_Usage");

	if(!m_b_mapped) {
		_ASSERTE(!b_Mapped()); // check the ogl state against m_b_mapped
		return 0;
	}

#ifdef _DEBUG
	int n_usage = m_n_usage; // must be bound
	glGetBufferParameterivARB(m_n_primary_target, GL_BUFFER_USAGE_ARB, &n_usage);
	_ASSERTE(n_usage == m_n_usage);
#endif

	return m_n_usage;
}

/*
 *	bool CVertexBufferObject::b_Mapped()
 *		- returns true in case buffer contents were mapped into client-space memory
 *		- it's possible to obtain pointer to buffer data by call to p_BufferPointer()
 */
bool CVertexBufferObject::b_Mapped()
{
	__FuncGuard("CVertexBufferObject::b_Mapped");

#ifdef _DEBUG
	int b_mapped = m_b_mapped; // must be bound
	glGetBufferParameterivARB(m_n_primary_target, GL_BUFFER_MAPPED_ARB, &b_mapped);
	_ASSERTE(m_b_mapped == (b_mapped == GL_TRUE));
#endif

	return m_b_mapped;
}

/*
 *	int CVertexBufferObject::n_AccessFlag()
 *		- returns access flag to buffer in case it has been mapped (one of GL_READ_ONLY_ARB,
 *		  GL_WRITE_ONLY_ARB, GL_READ_WRITE_ARB), in case buffer is not mapped, return 0
 */
int CVertexBufferObject::n_AccessFlag()
{
	__FuncGuard("CVertexBufferObject::n_AccessFlag");

	if(!m_b_mapped) {
		_ASSERTE(!b_Mapped()); // check the ogl state against m_b_mapped
		return 0;
	}

#ifdef _DEBUG
	int n_access = m_n_access; // must be bound
	glGetBufferParameterivARB(m_n_primary_target, GL_BUFFER_ACCESS_ARB, &n_access);
	_ASSERTE(m_n_access == n_access);
#endif

	return m_n_access;
}

/*
 *								=== ~CVertexBufferObject ===
 */

/*
 *								=== CPixelBufferObject ===
 */

/*
 *	CPixelBufferObject::CPixelBufferObject(GLenum n_target = GL_PIXEL_PACK_BUFFER_ARB)
 *		- default constructor, valid parameters are GL_ARRAY_BUFFER_ARB (vertex data),
 *		  GL_ELEMENT_ARRAY_BUFFER_ARB (vertex indices), GL_PIXEL_PACK_BUFFER_ARB
 *		  (pixel-buffer in mode for pixel storage) and GL_PIXEL_UNPACK_BUFFER_ARB
 *		  (pixel-buffer in mode for for reading pixels)
 */
CPixelBufferObject::CPixelBufferObject(GLenum n_target/* = GL_PIXEL_PACK_BUFFER_ARB*/)
	:CVertexBufferObject(n_target)
{}

/*
 *	static bool CPixelBufferObject::b_Supported()
 *		- returns true when GL_ARB_VERTEX_BUFFER_OBJECT and GL_ARB_PIXEL_BUFFER_OBJECT
 *		  extensions are supported and all entry points were found
 */
bool CPixelBufferObject::b_Supported()
{
	return CVertexBufferObject::b_Supported() &&
		CGLExtensionHandler::b_SupportedExtension("GL_ARB_pixel_buffer_object");
}

/*
 *								=== ~CPixelBufferObject ===
 */

/*
 *		-end-of-file-
 */
